% Creator: Andy Worsley
% Last Modified: 11/02/2008

% DESCRIPTION
% -----------
% Finds the loss parameters (efficiencies) for two LOOPYs by testing a
% series of loss parameters and minimising the least squares for each with 
% respect to the photon statistics using 'lsqnonneg' function and the
% assumption that the input state is a diagonal matrix (photon pairs).

% MAKES A FAT VECTOR OF ALL DATA?


% NOTES
% -----
% I have not found a a least squares algorithm for an equation of the form
% P = Q[rho]R, where all the variables are matrices, so the first part of 
% the program reduces the problem to a standard p = Mx form for a least 
% squares fit, where the components of x are the diagonals of [rho].


% DEPENDENCIES
% ------------
% Function "Loss" must be available - creates loss matrices
% Function "nck" must be available - nchoosek with error suppressed for high
% coefficients
% P_kn1 and P_kn2 must be created, and click must be available

% P_kn1: convolution matrix for LOOPY 1
% P_kn2: convolution matrix for LOOPY 2
% click: joint statistics count data


% OUTPUT
% ------
% losses_bestrn: pair of loss parameters with the minimum fit residual
%      x_bestrn: column vector of the fitted input state
%           rho: diagonal matrix of fitted input state




% -----
% START
% -----

% ---- INITIALISE AND PREALLOCATE VARIABLES ----
N = size(P_kn1,2) - 1; % Read highest considered photon pair number from convolution matrix size

n = 0; % Counting variables
m = 0;
prow = 0;
mrow = 0;
loopnum = 0;

l1 = 0; % Loss parameter in LOOPY 1
l2 = 0; % Loss parameter in LOOPY 2
L1 = zeros(N+1,N+1); % Loss matrix in LOOPY 1
L2 = zeros(N+1,N+1); % Loss matrix in LOOPY 2

Q = zeros(81,N+1); 
R = zeros(81,N+1);

p = zeros(81,1);
M = zeros(81,N+1);
flag = 0;

% Find statistics from click counts
click = Aall;
for r=1:30
    P(1:9,1:9,r) = click(:,:,r)/sum(sum(click(:,:,r)));
end

% ------------------------------------------------------------------------
% ITERATION PARAMETERS - MAY BE CHANGED BY THE USER TO OPTIMIZE ITERATION
% ------------------------------------------------------------------------
% The 'losswidths' vector specifies the width of the
% loss space to test for each iteration. The 'reslosses' vector specifies
% how many test points to take for each iteration. The more elements there
% are in the vector, the more iterations will be done.
% SIZE OF LOSSWIDTHS AND RESLOSSES MUST BE THE SAME

midloss1 = 0.50; % Puts centre of first iteration for loss parameter 1 at 0.5
midloss2 = 0.50; % Puts centre of first iteration for loss parameter 2 at 0.5
losswidths = [0.5, 0.1, 0.02, 0.005]; % eg [0.5, ...] means first iteration will go +/- 0.5
reslosses =  [50,  20,  20,   10   ]; % eg [50, ...] means first iteration will take 50 points in the width
% ------------------------------------------------------------------------

% Loss parameter variables
widthloss = 0;
lossres = 0;

numlosses = max(reslosses) + 1;
numwidths = size(losswidths,2);
lwcnt = 0;

lossparms = linspace(0,1,numlosses);
lossparms2 = linspace(0,1,numlosses);

lossspaceres = zeros(numlosses,numlosses,numwidths); % Space of possible loss combinations for each iteration

% Variables for providing user with information about how much of the
% program has been completed
countdone = 0;
percentdone = 0;
totalcount = sum((reslosses+ones(1,size(reslosses,2))).*(reslosses+ones(1,size(reslosses,2))));


% Create columns of vectors for storing test results
x = zeros(N+1,(numlosses)^2);
resnorm = zeros(1,numlosses^2);
losses = zeros(2,numlosses^2);

% Columns for calculating optima
minresnorm = 0;
minrnind = 0;

% Return variables
x_bestrn = zeros(N+1,1);
losses_bestrn = zeros(2,1);


% ---- END OF INITIALISATION ----


% Check P_kn1 and P_kn2 are the same size
if ((size(P_kn1,1)~=size(P_kn2,1)) + (size(P_kn1,2)~=size(P_kn2,2)))
    error('Convolution matrices must be the same size');
end

% Check losswidths and reslosses are the same size
if size(losswidths)~=size(reslosses)
    error('Losswidths and reslosses must be the same size');
end

% Check losswidths and reslosses are the same size
if max(losswidths)>0.5
    error('Cannot have a losswidth element greater than 0.5');
end

% Use all P elements to make p a FAT! (81*30)x1 vector
prow = 1;
for r = 1:30
    for n = 1:9
        for m = 1:9
            p(prow,1) = P(n,m,r);
            prow = prow + 1;
        end
    end
end

% ---- Loss Resolution Loop ----
while lwcnt < numwidths
    lwcnt = lwcnt + 1;
    widthloss = losswidths(lwcnt);
    lossres = reslosses(lwcnt);
    numlosses = lossres + 1;
    
    % Create array of loss parameters
    if (midloss1 - widthloss) < 0.0
        midloss1 = widthloss;
    end
    if (midloss2 - widthloss) < 0.0
        midloss2 = widthloss;
    end
    if (midloss1 + widthloss) > 1.0
        midloss1 = 1.0 - widthloss;
    end
    if (midloss2 + widthloss) > 1.0
        midloss2 = 1.0 - widthloss;
    end

    lossparms = linspace((midloss1 - widthloss),(midloss1 + widthloss), numlosses);
    lossparms2 = linspace((midloss2 - widthloss),(midloss2 + widthloss), numlosses);

    loss1count = 0;
    loss2count = 0;

    
    % ---- Loss 1 Loop ----
    loopnum = 1;
    for loss1count = 1:(numlosses)
        l1 = lossparms(loss1count); % Assign loss parameter
        L1 = Loss(l1, N); % Create loss matrix

        % ---- Loss 2 Loop ----
        for loss2count = 1:(numlosses)
            l2 = lossparms2(loss2count); % Assign loss parameter
            L2 = Loss(l2, N); % Create loss matrix

            % Calculate working matrices - initially need P = Q*rho*R
            Q = P_kn1*L1;
            R = L2'*P_kn2';       

            % Create 81x(N+1) M matrix - to convert to form p = Mx
            mrow = 1;
            for n = 1:9
                for m = 1:9
                M(mrow,:) = Q(n,:).*R(:,m)';    
                mrow = mrow + 1;
                end
            end
            
            % Replicate with a matrix * identity matrix.

            
            
            % Solve least squares problem to find photon statistics, x, and
            % record the loss parameters to which this corresponds
            %[x(:,loopnum), flag, resnorm(1,loopnum)] = lsqr(M,p,[],100,[],[],x_start);
            
            [x(:,loopnum), resnorm(1,loopnum)] = lsqnonneg(M,p);
            losses(:,loopnum) = [l1 ; l2];

            % Create array for resnorm & probdiff in loss space - creates data
            % for a 2-D map of the residual norm for visual inspection
            lossspaceres(loss1count,loss2count,lwcnt) = resnorm(1,loopnum);

            loopnum = loopnum + 1;
        end % ---- Loss 2 Loop ----
        
    countdone = countdone + numlosses;
    percentdone = 100*countdone/totalcount
    
    end % ---- Loss 1 Loop ----

    % Find minimum residual norm, then return statistics and location
    [minresnorm, minrnind] = min(resnorm(1,:));
    losses_bestrn(:,1) = losses(:,minrnind);
    
    % Adjust parameters for next loss resolution level
    midloss1 = losses_bestrn(1,1);
    midloss2 = losses_bestrn(2,1);
    
end
% ---- End of Loss Resolution Loop ----



% ---- Final Return of Data ----
losses_bestrn(:,1) = losses(:,minrnind);
L1 = Loss(losses_bestrn(1,1),N);
L2 = Loss(losses_bestrn(2,1),N);